//
// XMM Science Analysis System
//
// Giuseppe Vacanti (cosine) 
// June 2000
//
#include "TimeOrderedTable.h"
#include <errstr.h>
#include <Tracer.h>
#include <SasDalUtils.h>
#include <utility>
#include <algorithm>

static bool comPair(const pair<STime, unsigned>& rhs, const pair<STime, unsigned>& lhs)
{
  PROC
  RETURN rhs.first < lhs.first;
}

static bool equalPair(const pair<STime, unsigned>& rhs, const pair<STime, unsigned>& lhs)
{
  PROC
  RETURN rhs.second == lhs.second;
}
  

TimeOrderedTable::TimeOrderedTable(const Table *tab, const string& timename = "TIME")
  : _tab(tab),
    //    _stimes(tab->rows()),
    _col(tab->column(timename)->data())
    
{
  PROC
  STime::errorHandling = STime::EXCEPTION;
  using namespace SasDalUtils;
  for(unsigned i = 0; i < _tab->rows(); i++){
    try{
      STime st(getStringColumnRow(_col, i));
      _stimes.push_back(pair<STime, unsigned>(st,i));
    }
    catch(STime::InvalidTime e){
      errstr << "Encouneterd a time string that could not be translated to a valid time. Dataset = " << tab->parent()->qualifiedName() << " Row = " << i
	     << "This row will be ignored."
	     << Sas::warning("InvalidTimeString");
    }
  }
  // The table must be sorted in time ascending order.
  pair<STime, unsigned> prev = _stimes[0];
  for(unsigned i = 1; i < _stimes.size(); i++){
    if(prev.first > _stimes[i].first)
      errstr << tab->name() + " is not sorted in time ascending order at row " << i << Sas::error("NotATimeSortedTable");
    prev = _stimes[i];
  }  
  STime::errorHandling = STime::ERRORANDEXIT;
  ENDPROC
}

TimeOrderedTable::~TimeOrderedTable () { PROC delete _col; ENDPROC }

const Table& TimeOrderedTable::table() const
{
  PROC
  RETURN *_tab;
}

unsigned TimeOrderedTable::rows() const
{
  PROC
  RETURN _tab->rows();
}

unsigned TimeOrderedTable::validTimes() const
{
  PROC
  RETURN _stimes.size();
}

const STime& TimeOrderedTable::first_time() const
{
  PROC
  RETURN _stimes[0].first;
}

const STime& TimeOrderedTable::last_time() const
{
  PROC
  RETURN _stimes[_stimes.size() - 1].first;
}

// -----------------------------
// lower and upper bound methods
// -----------------------------
// Say we have: t0 t1 t2 t3 t4 t5 ... and t is between t1 and t2.
// I want lower_bound to return t2, I want upper bound to return
// t1. So that if I have ts between t1 and t2, and te between t4 and
// t5, I can get the range of rows to be in order to extract hk data
// as:  lower_bound(ts), upper_bound(te), and this gives me ts <=
// [t2,t4] <= te.
//
// These are the methods to use when trying to find the valid ranges
// for PERIODIC hk parameters.

unsigned TimeOrderedTable::lower_bound(const STime& t) const
{
  PROC
  CI b = ::lower_bound(_stimes.begin(), _stimes.end(),
		       make_pair(t,0), comPair);
  if( b == _stimes.end() ||
      b == _stimes.begin()) {
    errstr << "Looking for the first row such that " << t
	   << " <= time[row] failed. Available time range: "
	   << first_time() << " - " << last_time()
	   << Sas::message(Msg::AppLib, Msg::Verbose);
    ALTRETURNV(rows())
  }
  RETURN b->second;
}

unsigned TimeOrderedTable::upper_bound(const STime& t) const
{
  PROC
  CI b = ::upper_bound(_stimes.begin(), _stimes.end(),
		       make_pair(t,0), comPair);
  if(b == _stimes.end() ||
     b == _stimes.begin()) {
    errstr << "Looking for tha last row such that " << t
	   << " >= time[row] failed. Available time range: "
	   << first_time() << " - " << last_time()
	   << Sas::message(Msg::AppLib, Msg::Verbose);
    ALTRETURNV(rows())
  }
  RETURN (b - 1)->second;
}

// ----------------------------------------------------------------------

bool TimeOrderedTable::row_has_time(const unsigned row) const
{
  PROC
  RETURN std::binary_search(_stimes.begin(), _stimes.end(),
			    make_pair(STime("now"), row), equalPair);
}

// ---------------------------
// first and last less methods
// ---------------------------
// Say we have t0 t1 t2 t3 t4 t5 ... and t is between t2 and
// t3. first_less_than(t) returns t2, so as last_less_than(t). The
// difference is there is there are duplicate times: t0 t1 t1 t1 t1 t2
// t3 t4 ...


unsigned TimeOrderedTable::last_less_than(const STime& t) const
{
  PROC
  for(unsigned i = rows() - 1; i != 0; --i)
    if(row_has_time(i))
      if(time_of_row(i) <= t)
	ALTRETURNV(i)
  RETURN rows();
}

unsigned TimeOrderedTable::first_less_than(const STime& t) const
{
  PROC
  for(unsigned i = 0; i < rows(); ++i)
    if(row_has_time(i))
      if(time_of_row(i) <= t)
	ALTRETURNV(i)
  RETURN rows();
}

const STime& TimeOrderedTable::time_of_row(const unsigned row) const
{
  PROC
  if(!row_has_time(row))
    errstr << "Row = " << row << " in "
	   <<  _tab->parent()->qualifiedName() << " has no valid time."
	   << Sas::error("NoValidTime");
  RETURN _stimes[row].first;
}
